/*
 * Date (yymmdd):               100208
 * Written by:                  Dr. Francisco J. Romero-Campero
 * Additional contributions by: Jonathan Blakes
 * (in no particular order)     Dr. Hongqing Cao
 *                              Dr. Natalio Krasnogor
 *				Dr. Jamie Twycross
 * Funded by:                   EPSRC grant EP/E017215/1
 * Website:                     http://www.infobiotic.org/
 * Contact:                     nxk@cs.nott.ac.uk
 */

#include <string>


#include "mc2ModelChecker.h"
#include <pmodelchecker/formulaBNF.h>
#include <mcss/mcss.h>
#include <mcss/Compartment.h>
#include <fstream>
#include <stdio.h>

using namespace PMODELCHECKER;

void PMODELCHECKER::useMC2(MCParameters* MCParams)
{
    //Checking if the simulations in MC2 format are already available
    if(MCParams->simulations_generatedMC2 == "false")
    {
        //When the simulations are not available in MC2, we first check if they are available in HDF5 format
        if(MCParams->simulations_generatedHDF5 == "true")
        {
            if(fopen(MCParams->simulations_file_hdf5.c_str(),"r") != NULL)
            {
                //When the simulations are already available in HDF5 we generate them in MC2 format
                generateSimulationsFileMC2(MCParams->simulations_file_hdf5,MCParams->number_samples,MCParams->simulations_file_MC2);
            }
        }
        else if(MCParams->simulations_generatedHDF5 == "false")
        {
            //When the simulations in HDF5 are not available the simulations in
            //MC2 format are generated from the model
            //generateSimulationsFromModel(MCParams->mcss_params_file,MCParams->simulations_file_hdf5,MCParams->number_samples,MCParams->simulations_file_MC2);
            generateSimulationsFileMC2(MCParams->mcss_params_file,
                                       MCParams->simulations_file_MC2,
                                       MCParams->number_samples,
                                       MCParams->molecular_species,
                                       MCParams->positions);
        }
        else
        {
            cerr << "Unknown value for simulations_generated (true|false): " << MCParams->simulations_generatedHDF5 << std::endl;
            exit(EXIT_FAILURE);
        }
    }

    if(fopen(MCParams->simulations_file_MC2.c_str(),"r") != NULL)
    {
        //When the simulations are already available in MC2 format the properties are checked
        checkProperties(MCParams->pathMC2,MCParams->simulations_file_MC2,MCParams->temporal_formulas,MCParams->results_file);
    }
    else
    {
        cerr << "Simulation file in MC2 format does not exist" << std::endl;
        exit(EXIT_FAILURE);
    }

}


std::string PMODELCHECKER::obtainHead(std::string simulationsHDF5)
{
    //The command that outputs the head with [ to the file headFile.txt
    std::string command = "mcss-postprocess -l -t 1 " + simulationsHDF5 + " | head -n 1 > headFile.temp";

    //Execute the command invoking the system
    int output = system(command.c_str());

    //Generate the head from the one in the headFile.txt
    std::string head;
    std::string fileName = "headFile.temp";
    ifstream headFile(fileName.c_str());

    if (headFile.is_open())
    {
        std::string line;
        while (! headFile.eof() )
        {
            getline (headFile,line);
            //cout << line << endl;
            head += line;
        }
        headFile.close();

        //Remove temporary file headFile.temp
        if( remove("headFile.temp") != 0 )
        {
            cerr << "Unable to remove temporary file: headFile.temp" << std::endl;
        }

    }
    else
    {
        cerr << "Unable to open file " << fileName << std::endl;
        exit(EXIT_FAILURE);
    }

    //Replacing [ with (
    size_t pos = head.find("[");

    while(pos != head.npos)
    {
        head.replace(pos,1,"(");
        pos = head.find("[",pos++);
    }

    //Replacing ] with )
    pos = head.find("]");

    while(pos != head.npos)
    {
        head.replace(pos,1,")");
        pos = head.find("]",pos++);
    }

    return head;
}


void PMODELCHECKER::generateSimulationsFileMC2(std::string simulationsHDF5,unsigned long runs,std::string simulationsMC2)
{

    std::cout << "Generating simulations in MC2 format" << std::endl;

    //Generating the file in MC2 format containing the simulations
    std::string headInformation = obtainHead(simulationsHDF5);
    ofstream MC2file;

    //First simulation
    MC2file.open(simulationsMC2.c_str(),ios::out);
    MC2file << headInformation << std::endl;
    MC2file.close();

    //Generate command
    std::string command = "mcss-postprocess -p -l -t 1 " + simulationsHDF5 + " >> " + simulationsMC2;
    //std::cout << "Command to run: " << command << std::endl;
    int output1 = system(command.c_str());


    for(int r = 1; r <= runs; ++r)
    {
        MC2file.open(simulationsMC2.c_str(),ios::app);
        MC2file << std::endl;
        MC2file << headInformation << std::endl;
        MC2file.close();

        //Generate command
        std::string command = "mcss-postprocess -p -l -t " + somethingToString<int>(r) + " " + simulationsHDF5 + " >> " + simulationsMC2;

        int output2 = system(command.c_str());
    }
}

void PMODELCHECKER::generateSimulationsFileMC2(std::string simulationParameters,
                                               std::string simulationMC2,
                                               unsigned long numberRuns,
                                               std::string molecularSpecies,
                                               std::string compartments)
{
    unsigned long i,j,run;
    unsigned long numberCompartments;
    unsigned long numberSpecies;
    double timeStep;
    double simulatedTime;
    double maxTime;

    //Loading the parameters for the simulations
    SimulationParameters *parameters = new SimulationParameters();
    parameters->read(const_cast<char *>(simulationParameters.c_str()));
    parameters->setParameter("runs","1");
    parameters->setParameter("log_type", "none"); //chose not to save a hdf5 file
    maxTime = parameters->max_time;
    timeStep = parameters->log_interval;
  
    // create p-system from parameters specified in parameter file
    Psystem *psystem = new Psystem(*parameters);

    numberCompartments = psystem->getNumCompartments();
    numberSpecies = psystem->getNumSpecies();

    delete parameters;

    //File to output the simulations
    ofstream simulations;
    simulations.open(simulationMC2.c_str(),ios::out);

    //Molecular species to output
    std::vector<unsigned long> moleculeIndeces = moleculesToOutput(psystem,molecularSpecies);

    //Compartments to output
    std::vector<unsigned long> compartmentIndeces = compartmentsToOutput(psystem,compartments);

    //Initialising the simulated time
    simulatedTime = 0;

    std::cout << "Running mcss" << std::endl;

    //Generating the number of simulation runs specified
    for(run=0;run<numberRuns;++run)
    {
        //Writing head of the simulation
        Alphabet *alphabet = psystem->getAlphabet();
        std::cout << "Simulation number " << run+1 << " of " << numberRuns << std::endl;

        simulations << "time ";
        for(std::vector<unsigned long>::iterator p = compartmentIndeces.begin();
            p != compartmentIndeces.end();
            ++p)
        {
            Simulation::Compartment *comp = psystem->getCompartment(*p);
            std::string compName = comp->getName();
            unsigned long Xpos = comp->getXPosition();
            unsigned long Ypos = comp->getYPosition();

            for(std::vector<unsigned long>::iterator mol = moleculeIndeces.begin();
                mol != moleculeIndeces.end();
                ++mol)
            {
                std::string speciesID = alphabet->getName(*mol);
                std::string XposStr = somethingToString<unsigned long>(Xpos);
                std::string YposStr = somethingToString<unsigned long>(Ypos);
                speciesID += "_" + XposStr + "_" + YposStr + "_" + compName;
                simulations << speciesID << " ";
            }
        }

        simulations << std::endl;

        //One simulation run
        while(simulatedTime <= maxTime)
        {
            simulations << simulatedTime << " ";

            //Output current number of molecules
            for(std::vector<unsigned long>::iterator p = compartmentIndeces.begin();
                p != compartmentIndeces.end();
                ++p)
            {

                for(std::vector<unsigned long>::iterator mol = moleculeIndeces.begin();
                    mol != moleculeIndeces.end();
                    ++mol)
                {
                   simulations << psystem->getSpeciesLevel(*mol,*p) << " ";
                }

            }

            simulations << std::endl;

            //Update simulated time
            simulatedTime += timeStep;

            //Execute simulation according to the time step
            psystem->execute(simulatedTime);
        }
        psystem->resetSimulation();
        simulations << std::endl;
        simulatedTime = 0;
    }

    simulations.close();
}

void PMODELCHECKER::checkProperties(std::string pathMC2, std::string simulationsFileMC2, std::string temporalFormulas, std::string resultFile)
{
    //Generate the expanded formulas
    std::map<std::string,
         std::map<std::map<std::string,double>,
                  std::string> > formulas = generateFormulas(temporalFormulas,
                                                             "expandedFormulas.temp");

    //Generate the command
    std::string command = "mc2 stoch " + simulationsFileMC2 + " expandedFormulas.temp " + "--numtraces=100 -quiet > " + resultFile;
    //Running MC2
    std::cout << "Running MC2" << std::endl;
    int output = system(command.c_str());

    //Remove temporary file with the expanded formula
    if( remove("expandedFormulas.temp") != 0 )
    {
       cerr << "Unable to remove temporary file: expandedFormula.temp" << std::endl;
    }

    //Convert MC2 result to the same format as the PRISM result
    generatePRISMFormat(formulas, resultFile);
}

void PMODELCHECKER::generateSimulationsFromModel(std::string mcssParamsFile, std::string HDF5file, unsigned long runs, std::string simulationsFileMC2)
{
    //Generate command to call mcss
    std::string command = "mcss " + mcssParamsFile + " data_file=" + HDF5file + " runs=" + somethingToString<unsigned long>(runs);

    //Call mcss using system
    std::cout << "Running mcss" << std::endl;
    int output = system(command.c_str());

    //Generate the file in MC2 format from the HDF5 file
    generateSimulationsFileMC2(HDF5file,runs,simulationsFileMC2);
}

//std::map< std::map< std::string, double>, std::string>
std::map<std::string,
         std::map<std::map<std::string,double>,
                  std::string> > PMODELCHECKER::generateFormulas(std::string temporalFormulas, std::string expandedFormulas)
{
    //Get content of file
    std::string fileContent = fileToString(temporalFormulas);

    //Parse the file
    formulas_p formulasParser;
    tree_parse_info<> result = boost::spirit::classic::ast_parse(fileContent.c_str(),formulasParser,space_p);

    //Checking if parsing has been successful
    if(!result.full)
    {
        cerr << "Error parsing file " << temporalFormulas << " at:" << std::endl;
        std::cout << result.stop << std::endl;
        exit(EXIT_FAILURE);
    }

    //The abstract syntax tree associated with the parsing of the formulas
    iter_t astTree = result.trees.begin();

    std::map<std::string,
             std::map<std::map<std::string,double>,
                     std::string> > res = expandFormulas(astTree);

    //Write expanded formulas to a file
    ofstream fileOutput;
    fileOutput.open(expandedFormulas.c_str(),ios::out);

    for(std::map<std::string,
             std::map<std::map<std::string,double>,
                     std::string> >::iterator oriForm = res.begin();
        oriForm != res.end();
        ++oriForm)
    {
        std::map< std::map< std::string, double>, std::string> instantiations = oriForm->second;

        for(std::map< std::map< std::string, double>, std::string>::iterator form = instantiations.begin();
            form != instantiations.end();
            ++form)
        {
            fileOutput << form->second << std::endl;
        }

    }

    fileOutput.close();

    return res;

}


std::map<std::string,std::map<std::map<std::string,double>, std::string> > PMODELCHECKER::expandFormulas(iter_t astTree)
{

    std::map<std::string,
             std::map<std::map<std::string,double>,
             std::string> > expandedFormulas;

    //Going through each temporal formula specification
    for(iter_t form = astTree->children.begin(); form != astTree->children.end(); ++form)
    {
        //Temporal logic formula
        std::string formula(form->value.begin(),form->value.end());

        //Parameters associated with the formula
        std::map< std::string, std::vector<string> > formulaParameters;

        iter_t parameters = form->children.begin();

        //For each parameter we collect its name, lower and upper bound and incremental step
        for(iter_t param = parameters->children.begin(); param != parameters->children.end(); ++param )
        {
            std::string parameterName(param->value.begin(),param->value.end());

            std::vector<std::string> parameterValues;

            iter_t lowerBoundDef = param->children.begin();
            std::string lowerBound(lowerBoundDef->value.begin(),lowerBoundDef->value.end());
            parameterValues.push_back(lowerBound);

            iter_t stepDef = param->children.begin() + 1;
            std::string step(stepDef->value.begin(),stepDef->value.end());
            parameterValues.push_back(step);

            iter_t upperBoundDef = param->children.begin() + 2;
            std::string upperBound(upperBoundDef->value.begin(),upperBoundDef->value.end());
            parameterValues.push_back(upperBound);

            //Inserting the parameter
            formulaParameters[parameterName] = parameterValues;
        }

        std::map< std::map< std::string, double> , std::string > instantiatedFormula = instantiateFormula(formula, formulaParameters);

        //Adding the expanded formulas coming form the current formula to the final result
        expandedFormulas[formula] = instantiatedFormula;
    }

    return expandedFormulas;
}


std::map< std::map< std::string, double >, std::string > PMODELCHECKER::instantiateFormula(std::string formula,
                                std::map< std::string, std::vector<string> > formulaParameters)
{

    std::map<std::map< std::string, double>, std::string > accumulator;
    std::map<std::string,double> params;

    accumulator[params] = formula;

    for(std::map< std::string, std::vector<string> >::iterator param = formulaParameters.begin(); \
        param != formulaParameters.end();
        ++param)
    {
        std::string paramName = param->first;
        std::vector<std::string> paramValues = param->second;

        double lowerBound = fromStringTo<double>(paramValues[0]);
        double step = fromStringTo<double>(paramValues[1]);
        double upperBound = fromStringTo<double>(paramValues[2]);

        std::map< std::map< std::string, double>, std::string > instantiation;

        for(double i = lowerBound; i <= upperBound; i = i + step)
        {
            std::string value = somethingToString<double>(i);

            std::map< std::map< std::string, double>, std::string > oneValueInstantiation = replaceParameter(paramName,value,accumulator);

            instantiation.insert(oneValueInstantiation.begin(),oneValueInstantiation.end());
        }

        accumulator = instantiation;
    }

    return accumulator;

}

std::map< std::map< std::string, double>, std::string > PMODELCHECKER::replaceParameter(std::string parameterName,
                      std::string parameterValue,
                      std::map< std::map< std::string, double>, std::string > formulas)
{
    std::map< std::map< std::string, double>, std::string > results;

    for(std::map< std::map< std::string, double>, std::string >::iterator form = formulas.begin(); \
        form != formulas.end();
        ++form)
    {
        std::string formula = form->second;
        std::map< std::string, double > parameters = form->first;

        std::string instantiatedFormula = substituteString(formula,parameterName,parameterValue);

        parameters[parameterName] = fromStringTo<double>(parameterValue);

        results[parameters] = instantiatedFormula;
    }

    return results;

}


std::string PMODELCHECKER::substituteString(std::string formula, std::string parameterName, std::string parameterValue)
{

    std::string instantiatedFormula = formula;
    size_t pos = 0;
    size_t posReplace = instantiatedFormula.find(parameterName,pos);
    size_t length = parameterName.length();

    while(posReplace != instantiatedFormula.npos)
    {
        instantiatedFormula.replace(posReplace,length,parameterValue);
        posReplace = instantiatedFormula.find(parameterName,posReplace);
    }

    return instantiatedFormula;
}

//Unexpanded formulas
std::vector<std::string> PMODELCHECKER::getFormulas(std::string temporalFormulas)
{
    std::vector<std::string> unexpandedFormulas;

    //Get content of file
    std::string fileContent = fileToString(temporalFormulas);

    //Parse the file
    formulas_p formulasParser;
    tree_parse_info<> result = boost::spirit::classic::ast_parse(fileContent.c_str(),formulasParser,space_p);

    //Checking if parsing has been successful
    if(!result.full)
    {
        cerr << "Error parsing file " << temporalFormulas << " at:" << std::endl;
        std::cout << result.stop << std::endl;
        exit(EXIT_FAILURE);
    }

    //The abstract syntax tree associated with the parsing of the formulas
    iter_t astTree = result.trees.begin();

    //Going through each temporal formula specification
    for(iter_t form = astTree->children.begin(); form != astTree->children.end(); ++form)
    {
        //Temporal logic formula
        std::string formula(form->value.begin(),form->value.end());

        //Adding the unexpanded formula

        unexpandedFormulas.push_back(formula);
    }

    return unexpandedFormulas;
}

void PMODELCHECKER::generatePRISMFormat(std::map<std::string,
                                  std::map<std::map<std::string,double>,
                                           std::string> > formulas,
                         std::string resultFile)
{
    //Read in the string with the results (comma separated values)
    ifstream inputFile(resultFile.c_str());
    std::string output;
    getline(inputFile,output);
    inputFile.close();

    //Create the output vector
    std::vector<std::string> outputVector = createVectorCSV(output);

    //Open output file
    ofstream outputFile(resultFile.c_str(),ios::out);

    //Writing the results for each formula
    int resultPosition = 0;
    for(std::map<std::string,
                 std::map<std::map<std::string,double>,
                                   std::string> >::iterator form = formulas.begin();
        form != formulas.end();
        ++form)
    {
        //Write the formula that has been analysed
        outputFile << form->first << ":" << std::endl;

        //Generating the header for the output of each formula
        std::map<std::map<std::string,double>, std::string > formula = form->second;
        std::map<std::string, double> parameters = (formula.begin())->first;
        std::string parameterName;
        std::string fileHeader;
        for(std::map<std::string, double>::iterator par = parameters.begin();
            par != parameters.end();
            ++par)
        {
            parameterName = par->first;
            fileHeader += parameterName + " ";
        }

        fileHeader += "Result";

        outputFile << fileHeader << std::endl;

        //Write results
       for(std::map< std::map< std::string, double >, std::string >::iterator f = formula.begin();
           f != formula.end();
           ++f)
       {
           std::map< std::string, double > params = f->first;
           std::string parameterValue;
           for(std::map<std::string, double>::iterator par = params.begin();
               par != params.end();
               ++par)
           {
               parameterValue = somethingToString<double>(par->second);
               outputFile << parameterValue << " ";
           }

           outputFile << outputVector[resultPosition] << std::endl;
           ++resultPosition;
       }

        outputFile << std::endl;
    }

    outputFile.close();
}

std::vector<std::string> PMODELCHECKER::createVectorCSV(std::string output,
                                                        std::string token)
{
    std::vector<std::string> outputVector;

    size_t pos1 = 0;
    size_t pos2 = output.find(token);

    while( pos2 != output.npos)
    {
        std::string result = output.substr(pos1,pos2-pos1);
        outputVector.push_back(result);

        pos1 = pos2 + 1;
        pos2 = output.find(token,pos1);
    }

    std::string result = output.substr(pos1,pos2-pos1);
    outputVector.push_back(result);

    return outputVector;
}


std::vector<unsigned long> PMODELCHECKER::moleculesToOutput(Psystem *psystem, std::string molecularSpecies)
{
    std::vector<unsigned long> moleculeIndeces;
    Simulation::Alphabet *alphabet = psystem->getAlphabet();


    if(molecularSpecies == "all")
    {
        unsigned long numberSpecies = alphabet->size();
        for(unsigned long i = 0; i < numberSpecies; ++i)
        {
            moleculeIndeces.push_back(i);
        }
    }
    else
    {
        std::vector<std::string> moleculeNames = createVectorCSV(molecularSpecies);

        for(std::vector<std::string>::iterator molec = moleculeNames.begin();
            molec != moleculeNames.end();
            ++molec)
        {
            moleculeIndeces.push_back(alphabet->getSpeciesIndex(*molec));
        }

    }

    return moleculeIndeces;
}

std::vector<unsigned long> PMODELCHECKER::compartmentsToOutput(Psystem *psystem, std::string compartmentPositions)
{
    std::vector<unsigned long> compartmentIndeces;

    if(compartmentPositions == "all")
    {
        unsigned long numberCompartments = psystem->getNumCompartments();
        for(unsigned long i = 0; i < numberCompartments; ++i)
        {
            compartmentIndeces.push_back(i);
        }
    }
    else
    {
        std::vector<std::string> positions = createVectorCSV(compartmentPositions, ":");

        for(std::vector<std::string>::iterator pos = positions.begin();
            pos != positions.end();
            ++pos)
        {
            std::vector<std::string> coordinateStr = createVectorCSV(*pos);

            unsigned long coordinates[2];
            coordinates[0] = fromStringTo<unsigned long>(coordinateStr[0]);
            coordinates[1] = fromStringTo<unsigned long>(coordinateStr[1]);

            unsigned long numberCompartments = psystem->getNumCompartments();
            for(unsigned long i=0; i < numberCompartments; ++i)
            {
                const unsigned long *position = psystem->getCompartment(i)->getPosition();
                if(coordinates[0] == position[0] && coordinates[1] == position[1])
                {
                    compartmentIndeces.push_back(i);
                }
            }


            
        }

    }

    return compartmentIndeces;
}
