/*
 * Date (yymmdd):               090320
 * Version:                     0.2
 * Written by:                  Dr. Francisco J. Romero-Campero
 * Additional contributions by: Jonathan Blakes
 * (in no particular order)     Dr. Hongqing Cao
 *                              Dr. Natalio Krasnogor
 *				Dr. Jamie Twycross
 * Funded by:                   EPSRC grant EP/E017215/1
 * Website:                     http://www.infobiotic.org/
 * Contact:                     nxk@cs.nott.ac.uk
 */

#include <ecsb/LPPRule.h>


#include <pmodelchecker/prismmodel.h>

#include "prismmodel.h"

using namespace std;
using namespace ECSB;
using namespace PMODELCHECKER;


PrismModel::PrismModel()
{

}


PrismModel::PrismModel(const PrismModel& orig)
{
    this->type =orig.type;
    this->constantsDouble = orig.constantsDouble;
    this->rulesDescription = orig.rulesDescription;
    this->constantsInt = orig.constantsInt;
    this->modules = orig.modules;
}

PrismModel::~PrismModel()
{

}

//Constructor from a LPP system object
PrismModel::PrismModel(ECSB::LPPsystem Pmodel)
{
    //The type of our PRISM models will always be "stochastic".
    this->type = "stochastic";

    //From the LPP model we get the automata and the P systems distributed over them.
    //std::vector<ECSB::automaton> automata = (Pmodel.getCellularAutomata()).getAutomata();
    std::map<std::vector<long>,std::string> positions = Pmodel.getPositions().getPositions();
    std::map<std::string,ECSB::PSystem> Psystems = Pmodel.getPsystems();

    //Create the PRISM modules from automata and P systems
    createPRISMmodules(positions,Psystems);

    //Distribute the commands representing P systems rules in the different
    //modules
    distributeCommands(positions,Psystems);

    //Create the standard rewards associated with a P system model
    //createStandardRewards();

}

//Creation of the PRISM modules.
void PrismModel::createPRISMmodules(std::map<std::vector<long>,std::string> positions,
			       std::map<std::string,ECSB::PSystem> Psystems){

    for(std::map<std::vector<long>,std::string>::iterator cell = positions.begin();
        cell != positions.end();
        ++cell) 
    {
        std::vector<long> position = cell->first;
        ECSB::PSystem pSystem = Psystems[cell->second];

        std::map<std::string,ECSB::LPPCompartment> compartments = pSystem.getMembraneStructure();

        for(std::map<std::string,ECSB::LPPCompartment>::iterator comp = compartments.begin();
            comp != compartments.end();
            ++comp)
        {
            //The name of a module consists of the position, name of the P system and name of the compartment
            std::string moduleName;

            if(position.size() == 1)
                moduleName += somethingToString<int>(position[0]);
            else if (position.size() == 2)
                moduleName += somethingToString<int>(position[0]) + "_" + somethingToString<int>(position[1]);
            else if (position.size() == 3)
            {
                moduleName += somethingToString<int>(position[0]) + "_" + somethingToString<int>(position[1]);
                moduleName += "_" + somethingToString<int>(position[2]);
            }
            else
            {
                cerr << "Position has got more than three coordinates" << std::endl;
                exit(EXIT_FAILURE);
            }

            moduleName += "_" + pSystem.getName() + comp->first;

            PrismModule Pmodule(moduleName,pSystem,comp->second);
            this->modules[moduleName] = Pmodule;

        }
    }

}


//Creation of the PRISM modules.
void PrismModel::distributeCommands(std::map<std::vector<long>,std::string> positions,
			       std::map<std::string,ECSB::PSystem> Psystems)
{

    for(std::map<std::vector<long>,std::string>::iterator cell = positions.begin();
        cell != positions.end();
        ++cell)
    {
        std::vector<long> position = cell->first;

        //The P system associated with each position.
        ECSB::PSystem pSystem = Psystems[cell->second];

        //The P systems considered here are assumed to have a single membrane
        std::map<std::string,ECSB::LPPCompartment> compartments =
              pSystem.getMembraneStructure();

        for(std::map<std::string,ECSB::LPPCompartment>::iterator comp = compartments.begin();
            comp != compartments.end();
            ++comp)
        {
            std::string moduleName;

           if(position.size() == 1)
                moduleName += somethingToString<int>(position[0]);
            else if (position.size() == 2)
                moduleName += somethingToString<int>(position[0]) + "_" + somethingToString<int>(position[1]);
            else if (position.size() == 3)
            {
                moduleName += somethingToString<int>(position[0]) + "_" + somethingToString<int>(position[1]);
                moduleName += "_" + somethingToString<int>(position[2]);
            }
            else
            {
                cerr << "Position has got more than three coordinates" << std::endl;
                exit(EXIT_FAILURE);
            }

            moduleName += "_" + pSystem.getName() + comp->first;

            std::vector<ECSB::LPPRule> rules = (pSystem.getRuleSets())[comp->first].getRules();

            for(std::vector<ECSB::LPPRule>::iterator r = rules.begin();
                r!=rules.end();
                ++r)
            {
                //A constant is added to specify the stochastic constant associated
                //with each rule.
                std::string constantName = "c_" + r->getName() + "_" + moduleName;
                addConstantDouble(constantName,r->getConstant());

                //Depending on the type of rule the translation is done in a
                //different way.
                if(r->getType() == "boundary")
                {
                    PrismCommand command(0,moduleName,*r,position,positions);
                    (this->modules)[moduleName].addCommand(command);
                    (this->rulesDescription)[command.getIdentifier()] = command.getRule();
                }
                else if(r->getType() == "translocation")
                {
                    PrismCommand command1(1,moduleName,*r,position,positions);
                    (this->modules)[moduleName].addCommand(command1);

                    PrismCommand command2(2,moduleName,*r,position,positions);
                    std::string target = targetModuleName(*r,position,positions);
                    (this->modules)[target].addCommand(command2);
                    (this->rulesDescription)[command1.getIdentifier()] = command1.getRule();
                }
            } // for rules

        } // for compartments

    } // for positions

}


//Auxiliar to find target module compartment for inserting the commands
std::string PrismModel::targetModuleName(ECSB::LPPRule rule,
        std::vector<long> position,
        std::map<std::vector<long>,std::string> positions)
{
    //The position of the target automaton is computed by adding the rule vector
    //to the position of the current automaton.
    std::vector<long> targetPosition;
    std::vector<int> ruleVector = rule.getVector();

    int dim = position.size();
    for(int i = 0; i < dim; ++i)
        targetPosition.push_back(ruleVector[i] + position[i]);

    //Finding the target automaton
    for(std::map<std::vector<long>,std::string>::iterator aut = positions.begin();
        aut != positions.end();
        ++aut)
    {
        std::vector<long> candidatePosition = aut->first;
        if(vectorEqual(candidatePosition,targetPosition))
        {
            //Constructing the targetModuleName
            std::string moduleName;

            if(position.size() == 1)
                moduleName += somethingToString<int>(targetPosition[0]);
            else if (targetPosition.size() == 2)
                moduleName += somethingToString<int>(targetPosition[0]) + "_" + somethingToString<int>(targetPosition[1]);
            else if (targetPosition.size() == 3)
            {
                moduleName += somethingToString<int>(targetPosition[0]) + "_" + somethingToString<int>(targetPosition[1]);
                moduleName += "_" + somethingToString<int>(targetPosition[2]);
            }
            else
            {
                cerr << "Position has got more than three coordinates" << std::endl;
                exit(EXIT_FAILURE);
            }

            moduleName += "_" + aut->second;

            return moduleName;
        }
    }
    
    cerr << "No target module in rule " <<  rule.getName() << std::endl;
    exit(EXIT_FAILURE);
}


bool PrismModel::vectorEqual(std::vector<long> vector1, std::vector<long> vector2)
{
    if(vector1.size() != vector2.size())
    {
        cerr << "Vectors of different sizes";
        exit(EXIT_FAILURE);
    }

    for(int i = 0; i < vector1.size(); ++i)
    {
        if(vector1[i] != vector2[i])
            return FALSE;
    }

    return TRUE;
}

void PrismModel::addConstantDouble(std::string constantName, double value)
{
    (this->constantsDouble).insert(pair<std::string,double>(constantName,value));
}

void PrismModel::addConstantInt(std::string constantName)
{
    (this->constantsInt).push_back(constantName);
}

//Write
void PrismModel::write(std::string fileName)
{

  //Output file stream for PRISM model
  ofstream prismFile;

  //Open the file for PRISM model
  prismFile.open(fileName.c_str(), ios::out);

  //First the type of the model is specified
  prismFile << this->type << std::endl << std::endl;

  for(std::map<std::string,double>::iterator c = (this->constantsDouble).begin();
     c != (this->constantsDouble).end();
     ++c)
  {
      prismFile << "const double ";

      if(c->second <= 0)
      {
          prismFile << c->first << ";" << std::endl;
      }
      else
          prismFile << c->first << " = " << c->second << ";" <<std::endl;
  }

  for(std::map<std::string,PrismModule>::iterator mod = (this->modules).begin();
     mod != (this->modules).end();
     ++mod)
  {
      std::map<std::string,class PrismVariable> vars = (mod->second).getVariables();

      for(std::map<std::string,class PrismVariable>::iterator var = vars.begin();
            var != vars.end();
            ++var)
      {
          if((var->second).getLowerBound() == "")
          {
              prismFile << "const int ";
              prismFile << "lb_" << (var->second).getName() << ";\t\t\t";
              prismFile << "//Lower bound for molecule " << (var->second).getName() << std::endl;

              addConstantInt("lb_" + (var->second).getName());
          }

          if((var->second).getUpperBound() == "")
          {
              prismFile << "const int ";
              prismFile << "ub_" << (var->second).getName() << ";\t\t\t";
              prismFile << "// Upper bound for molecule " << (var->second).getName() << std::endl;

              addConstantInt("ub_" + (var->second).getName());
          }
      }

  }
  prismFile << std::endl << std::endl;


  //Write the modules
  for(std::map<std::string,PrismModule>::iterator nameMod = (this->modules).begin();
     nameMod != (this->modules).end();
     ++nameMod)
  {
      PrismModule mod = nameMod->second;
      prismFile << "module m_" << mod.getName() << std::endl;
      prismFile << std::endl;

      std::map<std::string,PrismVariable> prismVariables = mod.getVariables();

      for(std::map<std::string,PrismVariable>::iterator var = prismVariables.begin();
        var != prismVariables.end();
        ++var)

      {

          prismFile << "\t" << (var->second).getName() << " : ";

          prismFile << "[ lb_" << (var->second).getName() << " .. ";

          prismFile << "ub_" << (var->second).getName() << " ] ";

          prismFile << "init " << (var->second).getInitialValue() << ";" << std::endl;

      }

    prismFile << std::endl << std::endl;

    std::map<std::string,PrismCommand> prismCommands = mod.getCommands();

    for(std::map<std::string,PrismCommand>::iterator com = prismCommands.begin();
        com != prismCommands.end();
        ++com)
    {
        
        PrismCommand prismCommand = com->second;

        prismFile << "\t";

        std::string ruleDescription = prismCommand.getRule();

        prismFile << "// " << ruleDescription << std::endl;
        prismFile << "\t";

        //Write the commands associated with each module
        prismFile << "[" << prismCommand.getIdentifier() << "] " << prismCommand.getPredicate();

        prismFile << " -> " << std::endl;
        prismFile << "\t\t";
        prismFile << prismCommand.getRate() << " : " << std::endl;

        std::vector<std::string> commandUpdates = prismCommand.getUpdates();

        for(std::vector<std::string>::iterator up = commandUpdates.begin();
        up != commandUpdates.end();
        ++up)
        {
            prismFile << "\t\t\t";
            prismFile << "(" << *up;
            if(up != commandUpdates.end() - 1)
                prismFile << ") &" << std::endl;
            else
                prismFile << ") ;" << std::endl;
        }

        prismFile << std::endl;
    }


    prismFile << std::endl;

    prismFile << "endmodule" << std::endl;




      prismFile << std::endl;
  }

  prismFile << std::endl;

  //Write the constants used in the rewards
  prismFile << "const int molecule;" << std::endl;
  prismFile << "const int rule;" << std::endl;
  //prismFile << "const double T;" << std::endl << std::endl;

  //Write the rewards corresponding to the molecules
  prismFile << "rewards \"molecules\"" << std::endl;
  prismFile << "\tmolecule = 0 : 1;" << std::endl;

  int molec = 1;

  for(std::map<std::string,PrismModule>::iterator nameMod = (this->modules).begin();
     nameMod != (this->modules).end();
     ++nameMod)
  {
      PrismModule mod = nameMod->second;
      std::map<std::string,PrismVariable> prismVariables = mod.getVariables();

      for(std::map<std::string,PrismVariable>::iterator var = prismVariables.begin();
        var != prismVariables.end();
        ++var)

      {
          prismFile << "\tmolecule = " << molec;
          prismFile << " : " << (var->second).getName() << ";";
          prismFile << std::endl;
          ++molec;
      }
  }

  prismFile << "endrewards" << std::endl;


  //Write the rewards corresponding to the rules
  prismFile << "rewards \"rules\"" << std::endl;

  int rule = 1;

  for(std::map<std::string,PrismModule>::iterator nameMod = (this->modules).begin();
     nameMod != (this->modules).end();
     ++nameMod)
  {
      PrismModule mod = nameMod->second;

      std::map<std::string,PrismCommand> prismCommands = mod.getCommands();

      for(std::map<std::string,PrismCommand>::iterator com = prismCommands.begin();
        com != prismCommands.end();
        ++com)
      {
        PrismCommand prismCommand = com->second;

        prismFile << "\t[ " << prismCommand.getIdentifier() << " ] ";
        prismFile << "rule = " << rule << ": 1;" << std::endl;
        ++rule;
      }
  }

  prismFile << "endrewards" << std::endl;

  prismFile.close();

}

//Write variables in a XML file for communication with other tools
void PrismModel::writeParameters(std::string fileName)
{
    //Create the XML file containing the parameters
    xmlpp::Document parametersXML;
    xmlpp::Element* rootNode = parametersXML.create_root_node("params");

    //Paramters corresponding to the variables of the model (molecules in different compartments)
    xmlpp::Element* variablesNode = rootNode->add_child("modelVariables");

    int i = 1; //Varibale index for id

    for(std::map<std::string,PrismModule>::iterator module = this->modules.begin();
        module != this->modules.end();
        ++module)
    {
         std::map<std::string,PrismVariable> moduleVariables = (module->second).getVariables();

         for(std::map<std::string,PrismVariable>::iterator var = moduleVariables.begin();
            var != moduleVariables.end();
            ++var)
         {
             //new variable
             xmlpp::Element* varNode = variablesNode->add_child("variable");

             //variable id
             std::string varId = "v";
             varId += somethingToString<int>(i);
             varNode->set_attribute("id",varId);
             ++i;

             //variable name
             xmlpp::Element* varName = varNode->add_child("name");
             varName->add_child_text((var->second).getName());
             
             //variable description
             xmlpp::Element* varDescription = varNode->add_child("description");
             varDescription->add_child_text((var->second).getDescription());
         }        
    }

    //Parameters corresponding to the constants of the rules in the model
    xmlpp::Element* ruleConstantsNode = rootNode->add_child("ruleConstants");

    int j = 1; //Constant index for id
    for(std::map<std::string,double>::iterator constant = (this->constantsDouble).begin();
        constant != (this->constantsDouble).end();
        ++constant)
    {

      if(constant->second <= 0)
      {
        //new constant
        xmlpp::Element* constNode = ruleConstantsNode->add_child("constant");

        //constant id
        std::string constId = "c";
        constId += somethingToString<int>(j);
        constNode->set_attribute("id",constId);
        ++j;

        //constant name
        xmlpp::Element* constName = constNode->add_child("name");
        constName->add_child_text(constant->first);

        //constant description obtain from the name of a constant which consists of
        // c_{ruleName}_{posX}_{posY}_{compartmentName}
        xmlpp::Element* constDescription = constNode->add_child("description");

        //compartmentName
        size_t lastPos = (constant->first).rfind("_");
        std::string compartmentName = (constant->first).substr(lastPos + 1);

        //posY
        size_t positionY = (constant->first).rfind("_",lastPos-1);
        std::string posY = (constant->first).substr(positionY+1, lastPos - positionY - 1);

        //posX
        size_t positionX = (constant->first).rfind("_",positionY-1);
        std::string posX = (constant->first).substr(positionX+1, positionY - positionX -1);

        //rule name
        size_t firstPos = (constant->first).find("_");
        std::string ruleName = (constant->first).substr(firstPos+1);//, positionX - firstPos - 1);
        std::string ruleDescription = (this->rulesDescription)[ruleName];

        std::string description = "Constant from rule " + ruleName + " : " + ruleDescription;
        description += " in compartment " + compartmentName;
        description += " at position (" + posX + "," + posY + ")";


        constDescription->add_child_text(description);
      }

    }

    //Parameters corresponding to the constants of the rules in the model
    xmlpp::Element* moleculeConstantsNode = rootNode->add_child("moleculeConstants");

    int k = 1; //Constant index for id

    for(std::vector<std::string>::iterator molConstant = this->constantsInt.begin();
        molConstant != this->constantsInt.end();
        ++molConstant)
    {
        //new molecule constant
        xmlpp::Element* constMoleculeNode = moleculeConstantsNode->add_child("constant");

        //constant id
        std::string constMoleculeId = "c";
        constMoleculeId += somethingToString<int>(k);
        constMoleculeNode->set_attribute("id",constMoleculeId);
        ++k;

        //constant name
        xmlpp::Element* constMoleculeName = constMoleculeNode->add_child("name");
        constMoleculeName->add_child_text(*molConstant);

        //constant description obtain from the name of a constant which consists of
        //lb_{moleculeName}_{posX}_{posY}_{compartmentName}
        xmlpp::Element* constMoleculeDescription = constMoleculeNode->add_child("description");

        std::string description;

        //Distinguishing between lower and upper bound
        size_t firstPos = molConstant->find("_");
        std::string bound = molConstant->substr(0,firstPos);

        if(bound == "lb")
        {
            description = "Lower bound for ";
        }
        else if(bound == "ub")
        {
            description = "Upper bound for ";
        }
        else
        {
            cerr << "Not accepted constant name: " << *molConstant << std::endl;
            exit(EXIT_FAILURE);
        }

        //Molecule name
        size_t secondPos = molConstant->find("_",firstPos+1);
        std::string moleculeName = molConstant->substr(firstPos+1,secondPos - firstPos - 1);
        description += moleculeName;

        //Compartment name
        size_t lastPos = molConstant->rfind("_");
        std::string compartmentName = molConstant->substr(lastPos+1);
        description += " in compartment " + compartmentName;

        //Position
        size_t thirdPos = molConstant->find("_",secondPos+1);
        std::string posX = molConstant->substr(secondPos+1, thirdPos - secondPos - 1);
        std::string posY = molConstant->substr(thirdPos+1, lastPos - thirdPos - 1);

        description += " at position (" + posX + "," + posY + ")";
        
        constMoleculeDescription->add_child_text(description);

    }

    //Parameters corresponding to the rewards
    xmlpp::Element* rewardConstantsNode = rootNode->add_child("rewardConstants");

    int molec = 1; //Constant index for id

    for(std::map<std::string,PrismModule>::iterator nameMod = (this->modules).begin();
        nameMod != (this->modules).end();
        ++nameMod)
    {
        PrismModule mod = nameMod->second;
        std::map<std::string,PrismVariable> prismVariables = mod.getVariables();

        for(std::map<std::string,PrismVariable>::iterator var = prismVariables.begin();
            var != prismVariables.end();
            ++var)
        {
            //new molecule reward constant
            xmlpp::Element* rewardMoleculeNode = rewardConstantsNode->add_child("constant");

            //reward constant name
            xmlpp::Element* rewardMoleculeName = rewardMoleculeNode->add_child("name");
            rewardMoleculeName->add_child_text("molecule");

            //reward constant value
            xmlpp::Element* rewardMoleculeValue = rewardMoleculeNode->add_child("value");
            rewardMoleculeValue->add_child_text(somethingToString<int>(molec));

            //constant description obtain from the name of a constant which consists of
            //lb_{moleculeName}_{posX}_{posY}_{compartmentName}
            xmlpp::Element* rewardMoleculeDescription = rewardMoleculeNode->add_child("description");

            std::string description = "A reward corresponding to the number of molecules of ";
            description += (var->second).getName();
            description += " is associated with every state";
            rewardMoleculeDescription->add_child_text(description);


            molec++;
        }
  }






    //Writing the XML file with a formatted style
    parametersXML.write_to_file_formatted(fileName);

}


//Displayer
void PrismModel::show()
{
    //First the type of the model is specified
    std::cout << this->type << std::endl << std::endl;

    //Enumerating the constants of the model
    for(std::map<std::string,double>::iterator c = (this->constantsDouble).begin();
        c != (this->constantsDouble).end();
        ++c)
    {
        std::cout << "const double ";

        if(c->second < 0)
            std::cout << c->first << ";" << std::endl;
        else
            std::cout << c->first << " = " << c->second << ";" <<std::endl;
    }

    for(std::map<std::string,PrismModule>::iterator mod = (this->modules).begin();
        mod != (this->modules).end();
        ++mod)
    {
        std::map<std::string,class PrismVariable> vars = (mod->second).getVariables();
        for(std::map<std::string,class PrismVariable>::iterator var = vars.begin();
            var != vars.end();
            ++var)
        {
            if((var->second).getLowerBound() == "")
            {
                std::cout << "const int ";
                std::cout << "lb_" << (var->second).getName() << ";" << std::endl;
            }
  
            if((var->second).getUpperBound() == "")
            {
                std::cout << "const int ";
                std::cout << "ub_" << (var->second).getName() << ";" << std::endl;
            }
        }
        
    }

    std::cout << std::endl << std::endl;

    for(std::map<std::string,PrismModule>::iterator mod = (this->modules).begin();
        mod != (this->modules).end();
        ++mod)
    {
        (mod->second).show();
        std::cout << endl;
    }


}
