/*
 * Date (yymmdd):               090320
 * Version:                     0.2
 * Written by:                  Dr. Francisco J. Romero-Campero
 * Additional contributions by: Jonathan Blakes
 * (in no particular order)     Dr. Hongqing Cao
 *                              Dr. Natalio Krasnogor
 *				Dr. Jamie Twycross
 * Funded by:                   EPSRC grant EP/E017215/1
 * Website:                     http://www.infobiotic.org/
 * Contact:                     nxk@cs.nott.ac.uk
 */

#include <map>
#include <pmodelchecker/prismmodule.h>

using namespace ECSB;
using namespace PMODELCHECKER;

PrismModule::PrismModule() {
}

PrismModule::PrismModule(const PrismModule& orig)
{
    this->name = orig.name;
    this->variables = orig.variables;
    this->commands = orig.commands;
}

PrismModule::PrismModule(std::string moduleName,
                         ECSB::PSystem pSystem,
                         ECSB::LPPCompartment compartment)
{

    //Module name
    this->name = moduleName;

    //Module variables correspond to the objects in the P system
    std::vector<std::string> objects = pSystem.getAlphabet();

    //The variables initial value corresponds to the initial multiset of the
    //corresponding object
    std::map<std::string,InitialMultiset> initialMultisets =
                pSystem.getInitialMultisets();

    //The P systems considered here are assumed to have a single membrane
    std::map<std::string,ECSB::LPPCompartment> membranes =
            pSystem.getMembraneStructure();

    std::string label = compartment.getLabel();
    std::map<std::string,unsigned long> initialValues =
        (initialMultisets[label]).getMultiset();

    for(std::vector<std::string>::iterator obj = objects.begin();
        obj != objects.end();
        ++obj)
    {
        //generating the name of the variable
        std::string variableName = *obj + "_" + moduleName;
        PrismVariable var;

        var.setName(variableName);

        //assigning the initial value of the variable according to the initial multiplicity
        if(initialValues.find(*obj) != initialValues.end())
        {
            var.setInitialValue(somethingToString<unsigned long>(initialValues[*obj]));
        }
        else
        {
            var.setInitialValue("0");
        }

        //description of the variable

        std::string variableDescription = "Molecule " + *obj;

        //compartment name and position
        size_t pos = moduleName.find("_");
        pos = moduleName.find("_",pos+1);

        std::string compartmentName = moduleName.substr(pos+1);
        std::string position = moduleName.substr(0,pos);

        position.replace(position.find("_"),1,",");

        variableDescription += " in compartment " + compartmentName;

        variableDescription += " in position (" + position + ")" ;

        var.setDescription(variableDescription);

        this->variables[variableName] = var;
    }
}


/*
PrismModule::PrismModule(std::string moduleName, ECSB::PSystem pSystem)
{

    //Module name
    this->name = moduleName;

    //Module variables correspond to the objects in the P system
    std::vector<std::string> objects = pSystem.getAlphabet();

    //The variables initial value corresponds to the initial multiset of the
    //corresponding object
    std::map<std::string,InitialMultiset> initialMultisets =
                pSystem.getInitialMultisets();

    //The P systems considered here are assumed to have a single membrane
    std::map<std::string,ECSB::LPPCompartment> membranes =
            pSystem.getMembraneStructure();

    std::map<std::string,LPPCompartment>::iterator it = membranes.begin();
    std::string label = (it->second).getLabel();
    std::map<std::string,unsigned long> initialValues =
        (initialMultisets[label]).getMultiset();

    for(std::vector<std::string>::iterator obj = objects.begin();
        obj != objects.end();
        ++obj)
    {
        //generating the name of the variable
        std::string variableName = *obj + "_" + moduleName;
        PrismVariable var;
        
        var.setName(variableName);

        //assigning the initial value of the variable according to the initial multiplicity
        if(initialValues.find(*obj) != initialValues.end())
        {
            var.setInitialValue(somethingToString<unsigned long>(initialValues[*obj]));
        }
        else
        {
            var.setInitialValue("0");
        }

        //description of the variable

        std::string variableDescription = "Molecule " + *obj;

        //compartment name and position
        size_t pos = moduleName.find("_");
        pos = moduleName.find("_",pos+1);
        
        std::string compartmentName = moduleName.substr(pos+1);
        std::string position = moduleName.substr(0,pos);

        position.replace(position.find("_"),1,",");

        variableDescription += " in compartment " + compartmentName;

        variableDescription += " in position (" + position + ")" ;

        var.setDescription(variableDescription);

        (this->variables).insert(pair<std::string,PrismVariable>(variableName,var));
    }

   
}

 */

PrismModule::~PrismModule() {
}



std::string PrismModule::getName()
{
    return this->name;
}

std::map<std::string,PrismVariable> PrismModule::getVariables()
{
    return this->variables;
}

std::map<std::string,PrismCommand> PrismModule::getCommands()
{
    return this->commands;
}



void PrismModule::addCommand(PrismCommand command)
{
    (this->commands).insert(pair<std::string,PrismCommand>
            (command.getIdentifier(),command));

}


void PrismModule::write(ofstream prismFile)
{
    prismFile << "module m" << this->name << std::endl;

    prismFile << std::endl;

    for(std::map<std::string,PrismVariable>::iterator var = (this->variables).begin();
        var != (this->variables).end();
        ++var)
    {
        prismFile << "\t" << (var->second).getName() << " : ";
        prismFile << "[ lb_" << (var->second).getName() << " .. ";
        prismFile << "ub_" << (var->second).getName() << " ] ";
        prismFile << "init " << (var->second).getInitialValue() << ";" << std::endl;
    }

    prismFile << std::endl << std::endl;

    for(std::map<std::string,PrismCommand>::iterator com = (this->commands).begin();
        com != (this->commands).end();
        ++com)
    {

        std::cout << "\t";
//        (com->second).write(prismFile);
        prismFile << std::endl;
    }


    prismFile << std::endl;

    prismFile << "endmodule" << std::endl;

}

void PrismModule::show()
{
    std::cout << "module m" << this->name << std::endl;
    
    std::cout << endl;

    for(std::map<std::string,PrismVariable>::iterator var = (this->variables).begin();
        var != (this->variables).end();
        ++var)
    {
        std::cout << "\t" << (var->second).getName() << " : ";
        std::cout << "[ lb_" << (var->second).getName() << " .. ";
        std::cout << "ub_" << (var->second).getName() << " ] ";
        std::cout << "init " << (var->second).getInitialValue() << ";" << std::endl;
    }

    std::cout << endl << endl;

    for(std::map<std::string,PrismCommand>::iterator com = (this->commands).begin();
        com != (this->commands).end();
        ++com)
    {

        std::cout << "\t";
        (com->second).show();
        std::cout << std::endl;
    }


    std::cout << endl;

    std::cout << "endmodule" << std::endl;
}
